<?php

namespace Drupal\cms_content_sync\Plugin\cms_content_sync\field_handler;

use Drupal\gutenberg\Parser\BlockParser;
use Drupal\cms_content_sync\Plugin\FieldHandlerBase;
use Drupal\cms_content_sync\PullIntent;
use Drupal\cms_content_sync\PushIntent;
use Drupal\cms_content_sync\SyncIntent;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\StreamWrapper\PublicStream;

/**
 * Providing an implementation for formatted text. This handler pushes embedded
 * entities along like files or media entities in HTML.
 *
 * @FieldHandler(
 *   id = "cms_content_sync_default_formatted_text_handler",
 *   label = @Translation("Default Formatted Text"),
 *   weight = 90
 * )
 */
class DefaultFormattedTextHandler extends FieldHandlerBase {

  /**
   * {@inheritdoc}
   */
  public static function supports($entity_type, $bundle, $field_name, FieldDefinitionInterface $field) {
    $allowed = ['text_with_summary', 'text_long'];

    return FALSE !== in_array($field->getType(), $allowed);
  }

  /**
   * {@inheritdoc}
   */
  public function pull(PullIntent $intent) {
    $action = $intent->getAction();
    /**
     * @var \Drupal\Core\Entity\FieldableEntityInterface $entity
     */
    $entity = $intent->getEntity();

    // Deletion doesn't require any action on field basis for static data.
    if (SyncIntent::ACTION_DELETE == $action) {
      return FALSE;
    }

    if ($intent->shouldMergeChanges()) {
      return FALSE;
    }

    $data = $intent->getProperty($this->fieldName);

    if (empty($data)) {
      $entity->set($this->fieldName, NULL);
    }
    else {
      $result = [];

      foreach ($data as $item) {
        if (!empty($item['value'])) {
          // Replace node links correctly.
          $item['value'] = $this->replaceEntityReferenceLinks($intent, $item['value']);
        }
        $result[] = $item;
      }

      $entity->set($this->fieldName, $result);
    }

    return TRUE;
  }

  /**
   * Parse a text that uses regular HTML to look for dependencies like media or files.
   *
   * @param \Drupal\cms_content_sync\PushIntent $intent
   *   The PushIntent to add the dependencies to.
   * @param string $text
   *   The HTML to search.
   *
   * @return void
   */
  protected function extractDependenciesFromHtml(PushIntent $intent, string $text) {
    $add_file_uri_dependency = function ($matches) use ($intent) {
      // PDF files can have a #page=... anchor attached that we want to keep.
      $path = preg_replace('@#.*$@', '', $matches[1]);
      $uri = 'public://' . urldecode($path);

      /** @var FileInterface[] $files */
      $files = \Drupal::entityTypeManager()
        ->getStorage('file')
        ->loadByProperties(['uri' => $uri]);

      if (!count($files)) {
        \Drupal::logger('cms_content_sync')->error(
          'Failed to push referenced file by URI in the formatted text as the file is missing on this site: @uri<br>Flow: @flow_id | Pool: @pool_id',
          [
            '@uri' => $uri,
            '@flow_id' => $intent->getFlow()->id(),
            '@pool_id' => implode(',', $intent->getPoolIds()),
          ]
        );

        return '';
      }

      $file = reset($files);

      $intent->addDependency($file);

      return '';
    };

    $base_path = PublicStream::basePath();

    // Simple image embedding (default ckeditor + IMCE images)
    preg_replace_callback(
      '@<img[^>]+src="/' . $base_path . '/([^"]+)"@',
      $add_file_uri_dependency,
      $text
    );

    // Other file embedding (IMCE files)
    preg_replace_callback(
      '@<a[^>]+href="/' . $base_path . '/([^"]+)"@',
      $add_file_uri_dependency,
      $text
    );

    // Entity embedding (especially media)
    preg_replace_callback(
      '@<drupal-(entity|media)[^>]+data-entity-type="([^"]+)"\s+data-entity-uuid="([^"]+)"@',
      function ($matches) use ($intent) {
        $type = $matches[2];
        $uuid = $matches[3];

        $entity = \Drupal::service('entity.repository')->loadEntityByUuid($type, $uuid);

        if (!$entity) {
          return '';
        }

        $intent->addDependency($entity);

        return '';
      },
      $text
    );
  }

  /**
   * Parse a text that uses the 'gutenberg' filter in it's format to look for
   * dependencies like media or files.
   *
   * @param \Drupal\cms_content_sync\PushIntent $intent
   *   The PushIntent to add the dependencies to.
   * @param string $text
   *   The plain text as it's saved in the field.
   *
   * @return void
   */
  protected function extractDependenciesFromGutenberg(PushIntent $intent, string $text) {
    $block_parser = new BlockParser();
    $blocks = $block_parser->parse($text);

    foreach ($blocks as $block) {
      $this->extractDependenciesFromGutenbergBlock($intent, $block);
    }
  }

  /**
   * Extract the dependencies from the given block. The block is provided by the
   * Gutenberg module's BlockParser. Can be a string or an array with the block
   * definition.
   *
   * @param \Drupal\cms_content_sync\PushIntent $intent
   *   The PushIntent to add the dependencies to.
   * @param array|string $block
   *   The block to search for references.
   *
   * @return void
   */
  protected function extractDependenciesFromGutenbergBlock(PushIntent $intent, mixed $block) {
    if (is_string($block)) {
      $this->extractDependenciesFromHtml($intent, $block);
    }
    elseif (is_array($block)) {
      if (!empty($block["attrs"]["mediaAttrs"])) {
        $media = $block["attrs"]["mediaAttrs"];
        // core/image, core/file.
        if (!empty($media["data-entity-type"]) && !empty($media["data-entity-uuid"])) {
          $type = $media["data-entity-type"];
          $uuid = $media["data-entity-uuid"];
          $entity = \Drupal::service('entity.repository')->loadEntityByUuid($type, $uuid);
          if ($entity) {
            $intent->addDependency($entity);
          }
        }
        // core/gallery.
        elseif (is_array($media)) {
          foreach ($media as $item) {
            if (!empty($item["data-entity-type"]) && !empty($item["data-entity-uuid"])) {
              $type = $item["data-entity-type"];
              $uuid = $item["data-entity-uuid"];
              $entity = \Drupal::service('entity.repository')->loadEntityByUuid($type, $uuid);
              if ($entity) {
                $intent->addDependency($entity);
              }
            }
          }
        }
      }
      // drupalmedia/drupal-media-entity.
      if (!empty($block["attrs"]["mediaEntityIds"])) {
        foreach ($block["attrs"]["mediaEntityIds"] as $media_id) {
          $entity = \Drupal::service('entity.repository')->getActive('media', $media_id);
          if ($entity) {
            $intent->addDependency($entity);
          }
        }
      }
      if (!empty($block['innerHTML'])) {
        $this->extractDependenciesFromHtml($intent, $block["innerHTML"]);
      }
      if (!empty($block["innerBlocks"])) {
        foreach ($block["innerBlocks"] as $inner_block) {
          $this->extractDependenciesFromGutenbergBlock($intent, $inner_block);
        }
      }
      // Not handling drupalblock/* atm. Maybe useful if customers use custom
      // block content they want to sync with the module as dependencies as well.
    }
  }

  /**
   * Go through the text and check for referenced entities like media and files
   * to add them as dependencies to the sync so they are updated along with the
   * parent entity that contains this field.
   *
   * @param \Drupal\cms_content_sync\PushIntent $intent
   * @param string $text
   *   The field's ['value'].
   * @param string|null $format_id
   *   The field's ['format'].
   *
   * @return void
   */
  protected function extractDependencies(PushIntent $intent, string $text, ?string $format_id) {
    if (!$format_id) {
      $this->extractDependenciesFromHtml($intent, $text);
      return;
    }

    $formats = filter_formats();
    if (empty($formats[$format_id])) {
      return;
    }

    $format = $formats[$format_id];

    $filters = $format->filters()->getAll();

    if (!empty($filters['gutenberg']->status)) {
      $this->extractDependenciesFromGutenberg($intent, $text);
    }
    else {
      $this->extractDependenciesFromHtml($intent, $text);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function push(PushIntent $intent) {
    if (!parent::push($intent)) {
      return FALSE;
    }

    $entity = $intent->getEntity();

    foreach ($entity->get($this->fieldName)->getValue() as $item) {
      if (empty($item['value'])) {
        continue;
      }

      $this->extractDependencies($intent, $item['value'], $item['format'] ?? NULL);
    }

    return TRUE;
  }

  /**
   * Replace all "/node/..." links with their correct ID for the current site.
   *
   * @todo If a new entity is added, we should scan the database for existing
   * references to it that can now be resolved.
   *
   * @param \Drupal\cms_content_sync\PullIntent $intent
   * @param string $text
   *
   * @return string
   */
  protected function replaceEntityReferenceLinks($intent, $text) {
    $entity_repository = \Drupal::service('entity.repository');

    $replace_uri_callback = function ($matches) use ($intent) {
      $path = $matches[2];

      // PDF files can have a #page=... anchor attached that we want to keep.
      $anchor = NULL;
      if (FALSE !== strpos($path, '#')) {
        [$path, $anchor] = explode('#', $path);
      }

      // File URIs may include e.g. an ?itok=... from some text editors.
      $query = NULL;
      if (FALSE !== strpos($path, '?')) {
        [$path, $query] = explode('?', $path);
      }

      $parts = explode('/', $path);
      $prefix = '';
      if ($parts[0] === 'styles') {
        $prefix = $parts[0] . '/' . $parts[1] . '/';
        $parts = array_slice($parts, 2);
      }
      $file = NULL;
      $uri = NULL;
      for ($i = 0; $i < count($parts); ++$i) {
        $uri = 'public://' . urldecode(implode('/', array_slice($parts, $i)));

        /** @var FileInterface[] $files */
        $files = \Drupal::entityTypeManager()
          ->getStorage('file')
          ->loadByProperties(['uri' => $uri]);

        if (count($files)) {
          $file = reset($files);

          break;
        }

        $file = $intent->loadEmbeddedFile($uri);
        if (!empty($file)) {
          break;
        }
      }

      if (!$file) {
        \Drupal::logger('cms_content_sync')->error(
              'Failed to replace file URI in the formatted text as the file is missing on this site: @uri',
              [
                '@uri' => 'public://' . urldecode($path),
              ]
          );

        return $matches[1] . '"/404"';
      }

      $url = \Drupal::service('file_url_generator')->generateString($uri);

      if ($query) {
        $url .= '?' . $query;
      }
      if ($anchor) {
        $url .= '#' . $anchor;
      }

      return $matches[1] . '"' . /*$prefix .*/ $url . '"';
    };

    // Simple image embedding (default ckeditor + IMCE images)
    $text = preg_replace_callback(
          '@(<img[^>]+src=)"/sites/[^/]+/files/([^"]+)"@',
          $replace_uri_callback,
          $text
      );

    // Other file embedding (IMCE files)
    $text = preg_replace_callback(
          '@(<a[^>]+href=)"/sites/[^/]+/files/([^"]+)"@',
          $replace_uri_callback,
          $text
      );

    // Entity embedding (especially media)
    return preg_replace_callback(
          '@data-entity-uuid="([0-9a-z-]+)" href="/node/([0-9]+)"@',
          function ($matches) use ($entity_repository) {
              $uuid = $matches[1];
              $id = $matches[2];

            try {
                $node = $entity_repository->loadEntityByUuid('node', $uuid);
              if ($node) {
                $id = $node->id();
              }
            }
            catch (\Exception $e) {
            }

            return 'data-entity-uuid="' . $uuid . '" href="/node/' . $id . '"';
          },
          $text
      );
  }

}
